from typing import TYPE_CHECKING

from beanie.odm.utils.encoder import bson_encoder

if TYPE_CHECKING:
    from beanie.odm.documents import Document


def get_dict(document: "Document", to_db: bool = False):
    exclude = set()
    if document.id is None:
        exclude.add("_id")
    if not document.get_settings().model_settings.use_revision:
        exclude.add("revision_id")
    return bson_encoder.encode(
        document, by_alias=True, exclude=exclude, to_db=to_db
    )



import datetime
from collections import deque
from decimal import Decimal
from enum import Enum
from ipaddress import (
    IPv4Address,
    IPv4Interface,
    IPv4Network,
    IPv6Address,
    IPv6Interface,
    IPv6Network,
)
from pathlib import PurePath
from typing import Any, Callable, Dict, Type

from pydantic import SecretBytes, SecretStr
from pydantic.color import Color
from pydantic.json import isoformat

from beanie.odm.fields import Link

ENCODERS_BY_TYPE: Dict[Type[Any], Callable[[Any], Any]] = {
    Color: str,
    datetime.date: isoformat,
    datetime.time: isoformat,
    datetime.timedelta: lambda td: td.total_seconds(),
    Decimal: float,
    deque: list,
    IPv4Address: str,
    IPv4Interface: str,
    IPv4Network: str,
    IPv6Address: str,
    IPv6Interface: str,
    IPv6Network: str,
    SecretBytes: SecretBytes.get_secret_value,
    SecretStr: SecretStr.get_secret_value,
    Enum: lambda o: o.value,
    PurePath: str,
    Link: lambda l: l.ref,
}



from typing import Callable, Any


def get_class_path_for_method(f: Callable) -> str:
    return f"{f.__module__}.{f.__qualname__.split('.')[0]}"


def get_class_path_for_object(o: Any) -> str:
    return f"{o.__module__}.{o.__class__.__name__}"



from beanie.odm.fields import LinkTypes
from typing import TYPE_CHECKING, List, Dict, Any, Type

if TYPE_CHECKING:
    from beanie import Document


def construct_lookup_queries(cls: Type["Document"]) -> List[Dict[str, Any]]:
    queries = []
    link_fields = cls.get_link_fields()
    if link_fields is not None:
        for link_info in link_fields.values():
            queries.append(
                {
                    "$lookup": {
                        "from": link_info.model_class.get_motor_collection().name,  # type: ignore
                        "localField": f"{link_info.field}.$id",
                        "foreignField": "_id",
                        "as": link_info.field,
                    }
                }
            )
            if link_info.link_type == LinkTypes.DIRECT:
                queries.append({"$unwind": f"${link_info.field}"})  # type: ignore
    return queries



from collections import defaultdict
from datetime import datetime
from types import GeneratorType
from typing import (
    AbstractSet,
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    Tuple,
    Union,
)
from uuid import UUID

from bson import ObjectId, DBRef
from pydantic import BaseModel

from .bson import ENCODERS_BY_TYPE
from ..fields import LinkTypes


class Encoder:
    def __init__(self):
        self.encoders_by_class_tuples: Dict[
            Callable[[Any], Any], Tuple[Any, ...]
        ] = defaultdict(tuple)
        for type_, encoder in ENCODERS_BY_TYPE.items():
            self.encoders_by_class_tuples[encoder] += (type_,)

    def encode(
        self,
        obj: Any,
        exclude: Union[
            AbstractSet[Union[str, int]], Mapping[Union[str, int], Any], None
        ] = None,
        by_alias: bool = True,
        custom_encoder: Dict[Any, Callable[[Any], Any]] = None,
        to_db: bool = False,
    ) -> Any:
        from beanie.odm.documents import Document

        if custom_encoder is None:
            custom_encoder = {}
        if exclude is None:
            exclude = {}
        if not isinstance(exclude, (set, dict)):
            exclude = set(exclude)
        if isinstance(obj, Document):
            encoders = obj.get_settings().model_settings.bson_encoders
            if custom_encoder:
                encoders.update(custom_encoder)

            link_fields = obj.get_link_fields()
            obj_dict = {}
            for k, o in obj._iter(to_dict=False, by_alias=by_alias):
                if k not in exclude:  # TODO get exclude from the class
                    if link_fields and k in link_fields:
                        if link_fields[k].link_type == LinkTypes.LIST:
                            obj_dict[k] = [link.to_ref() for link in o]
                        if link_fields[k].link_type == LinkTypes.DIRECT:
                            obj_dict[k] = o.to_ref()
                    else:
                        obj_dict[k] = o
            return self.encode(obj_dict, custom_encoder=encoders, to_db=to_db)
        if isinstance(obj, BaseModel):
            encoders = {}
            if custom_encoder:
                encoders.update(custom_encoder)
            obj_dict = {}
            for k, o in obj._iter(to_dict=False, by_alias=by_alias):
                if k not in exclude:  # TODO get exclude from the class
                    obj_dict[k] = o

            return self.encode(obj_dict, custom_encoder=encoders, to_db=to_db)
        if isinstance(
            obj, (str, int, float, ObjectId, UUID, datetime, type(None), DBRef)
        ):
            return obj
        if isinstance(obj, dict):
            encoded_dict = {}
            for key, value in obj.items():
                encoded_value = self.encode(
                    value,
                    by_alias=by_alias,
                    custom_encoder=custom_encoder,
                    to_db=to_db,
                )
                encoded_dict[key] = encoded_value
            return encoded_dict
        if isinstance(obj, (list, set, frozenset, GeneratorType, tuple)):
            return [
                self.encode(
                    item,
                    exclude=exclude,
                    by_alias=by_alias,
                    custom_encoder=custom_encoder,
                    to_db=to_db,
                )
                for item in obj
            ]
        if custom_encoder:
            if type(obj) in custom_encoder:
                return custom_encoder[type(obj)](obj)
            for encoder_type, encoder in custom_encoder.items():
                if isinstance(obj, encoder_type):
                    return encoder(obj)
        if type(obj) in ENCODERS_BY_TYPE:
            return ENCODERS_BY_TYPE[type(obj)](obj)
        for encoder, classes_tuple in self.encoders_by_class_tuples.items():
            if isinstance(obj, classes_tuple):
                return encoder(obj)

        errors: List[Exception] = []
        try:
            data = dict(obj)
        except Exception as e:
            errors.append(e)
            try:
                data = vars(obj)
            except Exception as e:
                errors.append(e)
                raise ValueError(errors)
        return self.encode(
            data, by_alias=by_alias, custom_encoder=custom_encoder, to_db=to_db
        )


bson_encoder = Encoder()
